from typing import Callable, Iterable

from ...registry import registry
from .examples import SpanCoTExample, SpanExample
from .task import SpanTask
from .util import check_label_consistency, check_label_consistency_cot


@registry.llm_misc("spacy.LabelCheck.v1")
def make_label_check() -> Callable[[SpanTask], Iterable[SpanExample]]:
    return check_label_consistency


@registry.llm_misc("spacy.LabelCheckCoT.v1")
def make_label_check_cot() -> Callable[[SpanTask], Iterable[SpanCoTExample]]:
    return check_label_consistency_cot



from typing import Callable, Dict, List, Optional, Type, Union

from ...registry import registry
from ...ty import ExamplesConfigType, FewshotExample, Scorer, TaskResponseParser
from ...util import split_labels
from .parser import parse_responses_v1_v2_v3
from .task import DEFAULT_TEXTCAT_TEMPLATE_V1, DEFAULT_TEXTCAT_TEMPLATE_V2
from .task import DEFAULT_TEXTCAT_TEMPLATE_V3, TextCatTask
from .util import TextCatExample, score


@registry.llm_tasks("spacy.TextCat.v1")
def make_textcat_task(
    parse_responses: Optional[TaskResponseParser[TextCatTask]] = None,
    prompt_example_type: Optional[Type[FewshotExample]] = None,
    labels: str = "",
    examples: ExamplesConfigType = None,
    normalizer: Optional[Callable[[str], str]] = None,
    exclusive_classes: bool = False,
    allow_none: bool = True,
    verbose: bool = False,
    scorer: Optional[Scorer] = None,
) -> "TextCatTask":
    """TextCat.v1 task factory.

    You can use either binary or multilabel text classification based on the
    labels you provide.

    If a single label is provided, binary classification
    will be used. The label will get a score of `0` or `1` in `doc.cats`.

    Otherwise, multilabel classification will be used. The document labels
    in `doc.cats` will be a dictionary of strings and their score.

    Lastly, you can toggle between exclusive or no-exclusive text
    categorization by passing a flag to the `exclusive_classes` parameter.

    parse_responses (Optional[TaskResponseParser[TextCatTask]]): Callable for parsing LLM responses for this
        task.
    prompt_example_type (Optional[Type[FewshotExample]]): Type to use for fewshot examples.
    labels (str): Comma-separated list of labels to pass to the template.
        This task assumes binary classification if a single label is provided.
        Leave empty to populate it at initialization time (only if examples are provided).
    examples (Optional[Callable[[], Iterable[Any]]]): Optional callable that reads a file containing task examples for
        few-shot learning. If None is passed, then zero-shot learning will be used.
    normalizer (Optional[Callable[[str], str]]): Optional normalizer function.
    exclusive_classes (bool): If True, require the language model to suggest only one
        label per class. This is automatically set when using binary classification.
    allow_none (bool): if True, there might be cases where no label is applicable.
    verbose (bool): If True, show extra information.
    scorer (Optional[Scorer]): Scorer function.
    """
    labels_list = split_labels(labels)
    raw_examples = examples() if callable(examples) else examples
    example_type = prompt_example_type or TextCatExample
    textcat_examples = (
        [example_type(**eg) for eg in raw_examples] if raw_examples else None
    )
    return TextCatTask(
        parse_responses=parse_responses or parse_responses_v1_v2_v3,
        prompt_example_type=example_type,
        labels=labels_list,
        template=DEFAULT_TEXTCAT_TEMPLATE_V1,
        prompt_examples=textcat_examples,
        normalizer=normalizer,
        exclusive_classes=exclusive_classes,
        allow_none=allow_none,
        verbose=verbose,
        label_definitions=None,
        scorer=scorer or score,
    )


@registry.llm_tasks("spacy.TextCat.v2")
def make_textcat_task_v2(
    parse_responses: Optional[TaskResponseParser[TextCatTask]] = None,
    prompt_example_type: Optional[Type[FewshotExample]] = None,
    labels: Union[List[str], str] = [],
    template: str = DEFAULT_TEXTCAT_TEMPLATE_V2,
    examples: ExamplesConfigType = None,
    normalizer: Optional[Callable[[str], str]] = None,
    exclusive_classes: bool = False,
    allow_none: bool = True,
    verbose: bool = False,
    scorer: Optional[Scorer] = None,
) -> "TextCatTask":
    """TextCat.v2 task factory.

    You can use either binary or multilabel text classification based on the
    labels you provide.

    If a single label is provided, binary classification
    will be used. The label will get a score of `0` or `1` in `doc.cats`.

    Otherwise, multilabel classification will be used. The document labels
    in `doc.cats` will be a dictionary of strings and their score.

    Lastly, you can toggle between exclusive or no-exclusive text
    categorization by passing a flag to the `exclusive_classes` parameter.

    parse_responses (Optional[TaskResponseParser[TextCatTask]]): Callable for parsing LLM responses for this
        task.
    prompt_example_type (Optional[Type[FewshotExample]]): Type to use for fewshot examples.
    labels (Union[List[str], str]): List of labels to pass to the template,
        either an actual list or a comma-separated string.
        This task assumes binary classification if a single label is provided.
        Leave empty to populate it at initialization time (only if examples are provided).
    template (str): Prompt template passed to the model.
    examples (Optional[Callable[[], Iterable[Any]]]): Optional callable that reads a file containing task examples for
        few-shot learning. If None is passed, then zero-shot learning will be used.
    normalizer (Optional[Callable[[str], str]]): Optional normalizer function.
    exclusive_classes (bool): If True, require the language model to suggest only one
        label per class. This is automatically set when using binary classification.
    allow_none (bool): if True, there might be cases where no label is applicable.
    verbose (bool): If True, show extra information.
    scorer (Optional[Scorer]): Scorer function.
    """
    labels_list = split_labels(labels)
    raw_examples = examples() if callable(examples) else examples
    example_type = prompt_example_type or TextCatExample
    textcat_examples = (
        [example_type(**eg) for eg in raw_examples] if raw_examples else None
    )

    return TextCatTask(
        parse_responses=parse_responses or parse_responses_v1_v2_v3,
        prompt_example_type=example_type,
        labels=labels_list,
        template=template,
        prompt_examples=textcat_examples,
        normalizer=normalizer,
        exclusive_classes=exclusive_classes,
        allow_none=allow_none,
        verbose=verbose,
        label_definitions=None,
        scorer=scorer or score,
    )


@registry.llm_tasks("spacy.TextCat.v3")
def make_textcat_task_v3(
    parse_responses: Optional[TaskResponseParser[TextCatTask]] = None,
    prompt_example_type: Optional[Type[FewshotExample]] = None,
    labels: Union[List[str], str] = [],
    template: str = DEFAULT_TEXTCAT_TEMPLATE_V3,
    label_definitions: Optional[Dict[str, str]] = None,
    examples: ExamplesConfigType = None,
    normalizer: Optional[Callable[[str], str]] = None,
    exclusive_classes: bool = False,
    allow_none: bool = True,
    verbose: bool = False,
    scorer: Optional[Scorer] = None,
) -> "TextCatTask":
    """TextCat.v3 task factory.

    You can use either binary or multilabel text classification based on the
    labels you provide.

    If a single label is provided, binary classification
    will be used. The label will get a score of `0` or `1` in `doc.cats`.

    Otherwise, multilabel classification will be used. The document labels
    in `doc.cats` will be a dictionary of strings and their score.

    Lastly, you can toggle between exclusive or no-exclusive text
    categorization by passing a flag to the `exclusive_classes` parameter.

    parse_responses (Optional[TaskResponseParser[TextCatTask]]): Callable for parsing LLM responses for this
        task.
    prompt_example_type (Optional[Type[FewshotExample]]): Type to use for fewshot examples.
    labels (Union[List[str], str]): List of labels to pass to the template,
        either an actual list or a comma-separated string.
        This task assumes binary classification if a single label is provided.
        Leave empty to populate it at initialization time (only if examples are provided).
    template (str): Prompt template passed to the model.
    label_definitions (Optional[Dict[str, str]]): Optional dict mapping a label to a description of that label.
        These descriptions are added to the prompt to help instruct the LLM on what to extract.
    examples (Optional[Callable[[], Iterable[Any]]]): Optional callable that reads a file containing task examples for
        few-shot learning. If None is passed, then zero-shot learning will be used.
    normalizer (Optional[Callable[[str], str]]): Optional normalizer function.
    exclusive_classes (bool): If True, require the language model to suggest only one
        label per class. This is automatically set when using binary classification.
    allow_none (bool): if True, there might be cases where no label is applicable.
    verbose (bool): If True, show extra information.
    scorer (Optional[Scorer]): Scorer function.
    """

    labels_list = split_labels(labels)
    raw_examples = examples() if callable(examples) else examples
    example_type = prompt_example_type or TextCatExample
    textcat_examples = (
        [example_type(**eg) for eg in raw_examples] if raw_examples else None
    )

    return TextCatTask(
        parse_responses=parse_responses or parse_responses_v1_v2_v3,
        prompt_example_type=example_type,
        labels=labels_list,
        template=template,
        label_definitions=label_definitions,
        prompt_examples=textcat_examples,
        normalizer=normalizer,
        exclusive_classes=exclusive_classes,
        allow_none=allow_none,
        verbose=verbose,
        scorer=scorer or score,
    )



from collections import defaultdict
from typing import Any, Dict, Iterable

from spacy.scorer import get_ner_prf
from spacy.training import Example

from ...compat import Self
from ..span import SpanExample
from ..span.examples import SpanCoTExample


class NERExample(SpanExample):
    @classmethod
    def generate(cls, example: Example, **kwargs) -> Self:
        entities = defaultdict(list)
        for ent in example.reference.ents:
            entities[ent.label_].append(ent.text)

        return cls(text=example.reference.text, entities=entities)


class NERCoTExample(SpanCoTExample):
    @classmethod
    def generate(cls, example: Example, **kwargs) -> Self:
        return cls(
            text=example.reference.text,
            spans=SpanCoTExample._extract_span_reasons(example.reference.ents),
        )


def score(examples: Iterable[Example], **kwargs) -> Dict[str, Any]:
    """Score NER accuracy in examples.
    examples (Iterable[Example]): Examples to score.
    RETURNS (Dict[str, Any]): Dict with metric name -> score.
    """
    return get_ner_prf(examples)



# mypy: ignore-errors
import re

import pytest

from spacy_llm.models.rest.anthropic import Anthropic, Endpoints

from ..compat import has_anthropic_key


@pytest.mark.external
@pytest.mark.skipif(has_anthropic_key is False, reason="Anthropic API key unavailable")
def test_anthropic_api_response_is_correct():
    """Check if we're getting the expected response and we're parsing it properly"""
    anthropic = Anthropic(
        name="claude-instant-1",
        endpoint=Endpoints.COMPLETIONS.value,
        config={"max_tokens_to_sample": 10},
        strict=False,
        max_tries=10,
        interval=5.0,
        max_request_time=20,
    )

    prompt = "Count the number of characters in this string: hello"
    num_prompts = 3
    responses = anthropic(prompts=[prompt] * num_prompts)
    for response in responses:
        assert isinstance(response, str)


@pytest.mark.external
@pytest.mark.skipif(has_anthropic_key is False, reason="Anthropic API key unavailable")
def test_anthropic_api_response_when_error():
    """Check if error message shows up properly given incorrect config"""
    # Incorrect config c.f. https://console.anthropic.com/docs/api/reference
    incorrect_temperature = "one"  # should be an int
    with pytest.raises(ValueError, match="Request to Anthropic API failed:"):
        Anthropic(
            name="claude-instant-1",
            endpoint=Endpoints.COMPLETIONS.value,
            config={
                "max_tokens_to_sample": 10,
                "temperature": incorrect_temperature,
            },
            strict=False,
            max_tries=10,
            interval=5.0,
            max_request_time=20,
        )


@pytest.mark.external
@pytest.mark.skipif(has_anthropic_key is False, reason="Anthropic API key unavailable")
def test_anthropic_error_unsupported_model():
    """Ensure graceful handling of error when model is not supported"""
    incorrect_model = "x-gpt-3.5-turbo"
    with pytest.raises(
        ValueError, match=re.escape("Model 'x-gpt-3.5-turbo' is not supported")
    ):
        Anthropic(
            name=incorrect_model,
            endpoint=Endpoints.COMPLETIONS.value,
            config={"max_tokens_to_sample": 10},
            strict=False,
            max_tries=10,
            interval=5.0,
            max_request_time=20,
        )



from .normalizer import lowercase_normalizer, strip_normalizer
from .reader import fewshot_reader, file_reader
from .util import registry

__all__ = [
    "lowercase_normalizer",
    "strip_normalizer",
    "fewshot_reader",
    "file_reader",
    "registry",
]



from .model import Anthropic, Endpoints
from .registry import anthropic_claude_1, anthropic_claude_1_0, anthropic_claude_1_2
from .registry import anthropic_claude_1_3, anthropic_claude_instant_1
from .registry import anthropic_claude_instant_1_1

__all__ = [
    "Anthropic",
    "Endpoints",
    "anthropic_claude_1",
    "anthropic_claude_1_0",
    "anthropic_claude_1_2",
    "anthropic_claude_1_3",
    "anthropic_claude_instant_1",
    "anthropic_claude_instant_1_1",
]



from typing import Any, Callable, Dict, Iterable, Optional, Type

from confection import SimpleFrozenDict

from ...compat import has_langchain, langchain
from ...registry import registry

try:
    from langchain import base_language  # noqa: F401
    from langchain import llms  # noqa: F401
except (ImportError, AttributeError):
    llms = None


class LangChain:
    def __init__(
        self,
        name: str,
        api: str,
        config: Dict[Any, Any],
        query: Callable[
            ["langchain.base_language.BaseLanguageModel", Iterable[Any]],
            Iterable[Any],
        ],
    ):
        """Initializes model instance for integration APIs.
        name (str): Name of LangChain model to instantiate.
        api (str): Name of class/API.
        config (Dict[Any, Any]): Config passed on to LangChain model.
        query (Callable[[Any, Iterable[_PromptType]], Iterable[_ResponseType]]): Callable executing LLM prompts when
            supplied with the `integration` object.
        """
        self._langchain_model = LangChain.get_type_to_cls_dict()[api](
            model_name=name, **config
        )
        self.query = query
        self._check_installation()

    @staticmethod
    def get_type_to_cls_dict() -> Dict[
        str, Type["langchain.base_language.BaseLanguageModel"]
    ]:
        """Returns langchain.llms.type_to_cls_dict.
        RETURNS (Dict[str, Type[langchain.base_language.BaseLanguageModel]]): langchain.llms.type_to_cls_dict.
        """
        return langchain.llms.type_to_cls_dict

    def __call__(self, prompts: Iterable[Any]) -> Iterable[Any]:
        """Executes prompts on specified API.
        prompts (Iterable[Any]): Prompts to execute.
        RETURNS (Iterable[Any]): API responses.
        """
        return self.query(self._langchain_model, prompts)

    @staticmethod
    def query_langchain(
        model: "langchain.base_language.BaseLanguageModel", prompts: Iterable[Any]
    ) -> Iterable[Any]:
        """Query LangChain model naively.
        model (langchain.base_language.BaseLanguageModel): LangChain model.
        prompts (Iterable[Any]): Prompts to execute.
        RETURNS (Iterable[Any]): LLM responses.
        """
        return [model(pr) for pr in prompts]

    @staticmethod
    def _check_installation() -> None:
        """Checks whether `langchain` is installed. Raises an error otherwise."""
        if not has_langchain:
            raise ValueError(
                "The LangChain model requires `langchain` to be installed, which it is not. See "
                "https://github.com/hwchase17/langchain for installation instructions."
            )

    @staticmethod
    def _langchain_model_maker(class_id: str):
        def langchain_model(
            name: str,
            query: Optional[
                Callable[
                    ["langchain.base_language.BaseLanguageModel", Iterable[str]],
                    Iterable[str],
                ]
            ] = None,
            config: Dict[Any, Any] = SimpleFrozenDict(),
            langchain_class_id: str = class_id,
        ) -> Optional[Callable[[Iterable[Any]], Iterable[Any]]]:
            try:
                return LangChain(
                    name=name,
                    api=langchain_class_id,
                    config=config,
                    query=query_langchain() if query is None else query,
                )
            except ImportError as err:
                raise ValueError(
                    f"Failed to instantiate LangChain model {langchain_class_id}. Ensure all necessary dependencies "
                    f"are installed."
                ) from err

        return langchain_model

    @staticmethod
    def register_models() -> None:
        """Registers APIs supported by langchain (one API is registered as one model).
        Doesn't attempt to register anything if LangChain isn't installed or at least one LangChain model has been
        registered already.
        """
        if not has_langchain or any(
            [
                (handle.startswith("langchain"))
                for handle in registry.llm_models.get_all()
            ]
        ):
            return

        for class_id, cls in LangChain.get_type_to_cls_dict().items():
            registry.llm_models.register(
                f"langchain.{cls.__name__}.v1",
                func=LangChain._langchain_model_maker(class_id=class_id),
            )


@registry.llm_queries("spacy.CallLangChain.v1")
def query_langchain() -> (
    Callable[
        ["langchain.base_language.BaseLanguageModel", Iterable[Any]], Iterable[Any]
    ]
):
    """Returns query Callable for LangChain.
    RETURNS (Callable[["langchain.llms.BaseLLM", Iterable[Any]], Iterable[Any]]:): Callable executing simple prompts on
        the specified LangChain model.
    """
    return LangChain.query_langchain



from typing import Any, Callable, Dict, Iterable

from confection import SimpleFrozenDict

from ....registry import registry
from .model import AzureOpenAI, ModelType

_DEFAULT_TEMPERATURE = 0.0


@registry.llm_models("spacy.Azure.v1")
def azure_openai(
    name: str,
    base_url: str,
    model_type: ModelType,
    config: Dict[Any, Any] = SimpleFrozenDict(temperature=_DEFAULT_TEMPERATURE),
    strict: bool = AzureOpenAI.DEFAULT_STRICT,
    max_tries: int = AzureOpenAI.DEFAULT_MAX_TRIES,
    interval: float = AzureOpenAI.DEFAULT_INTERVAL,
    max_request_time: float = AzureOpenAI.DEFAULT_MAX_REQUEST_TIME,
    api_version: str = "2023-05-15",
) -> Callable[[Iterable[str]], Iterable[str]]:
    """Returns OpenAI instance for 'gpt-4' model using REST to prompt API.

    config (Dict[Any, Any]): LLM config passed on to the model's initialization.
    name (str): Name of the deployment to use. Note that this does not necessarily equal the name of the model used by
        that deployment, as deployment names in Azure OpenAI can be arbitrary.
    endpoint (str): The URL for your Azure OpenAI endpoint. This is usually something like
        "https://{prefix}.openai.azure.com/".
    model_type (ModelType): Whether the deployed model is a text completetion model (e. g.
        text-davinci-003) or a chat model (e. g. gpt-4).
    strict (bool): If True, ValueError is raised if the LLM API returns a malformed response (i. e. any kind of JSON
        or other response object that does not conform to the expectation of how a well-formed response object from
        this API should look like). If False, the API error responses are returned by __call__(), but no error will
        be raised.
    max_tries (int): Max. number of tries for API request.
    interval (float): Time interval (in seconds) for API retries in seconds. We implement a base 2 exponential backoff
        at each retry.
    max_request_time (float): Max. time (in seconds) to wait for request to terminate before raising an exception.
    api_version (str): API version to use.
    RETURNS (Callable[[Iterable[str]], Iterable[str]]]): OpenAI instance for 'gpt-4' model

    DOCS: https://spacy.io/api/large-language-models#models
    """
    return AzureOpenAI(
        name=name,
        endpoint=base_url,
        config=config,
        strict=strict,
        max_tries=max_tries,
        interval=interval,
        max_request_time=max_request_time,
        api_version=api_version,
        model_type=model_type,
    )
